/*
 * atanks - obliterate each other with oversize weapons
 * Copyright (C) 2003  Thomas Hudson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * */

/*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
fade.cc

Provides graphical transitions via change and quickChange, as prototyped in
main.h
@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*/

/*
TODO
 + Add more faders
 + Improve/complete documentation
*/



#include "main.h"



/*****************************************************************************
Internal fades

The following section of code defines the internal fade routines.  These
routines are used by the change function to perform transitions.  To add a
new fader, simply implement a function as seen below and then add it to
change's function pointer array.
*****************************************************************************/

static void spiralPixelFade (BITMAP *target) {
	int blockSize = 8;
	int spiralSize = 20;
	int length = spiralSize, x = 0, y = 0;
	int stepSize = (spiralSize + 1) * blockSize;
	int x2, y2;
	int count;

	for (count = 0, y = stepSize - blockSize; count < length; count++, y -= blockSize)
		for (x2 = 0; x2 <= SCREEN_WIDTH; x2 += stepSize)
			for (y2 = 0; y2 <= SCREEN_HEIGHT; y2 += stepSize)
				blit (target, screen, x2+x, y2+y, x2+x, y2+y, blockSize, blockSize);
	while (length) {
		for (count = 0; count < length; count++, x += blockSize)
			for (x2 = 0; x2 <= SCREEN_WIDTH; x2 += stepSize)
				for (y2 = 0; y2 <= SCREEN_HEIGHT; y2 += stepSize)
					blit (target, screen, x2+x, y2+y, x2+x, y2+y, blockSize, blockSize);
		for (count = 0; count < length; count++, y += blockSize)
			for (x2 = 0; x2 <= SCREEN_WIDTH; x2 += stepSize)
				for (y2 = 0; y2 <= SCREEN_HEIGHT; y2 += stepSize)
					blit (target, screen, x2+x, y2+y, x2+x, y2+y, blockSize, blockSize);
		length--;
		for (count = 0; count < length; count++, x -= blockSize)
			for (x2 = 0; x2 <= SCREEN_WIDTH; x2 += stepSize)
				for (y2 = 0; y2 <= SCREEN_HEIGHT; y2 += stepSize)
					blit (target, screen, x2+x, y2+y, x2+x, y2+y, blockSize, blockSize);
		for (count = 0; count < length; count++, y -= blockSize)
			for (x2 = 0; x2 <= SCREEN_WIDTH; x2 += stepSize)
				for (y2 = 0; y2 <= SCREEN_HEIGHT; y2 += stepSize)
					blit (target, screen, x2+x, y2+y, x2+x, y2+y, blockSize, blockSize);
		length--;
	}
	for (x2 = 0; x2 <= SCREEN_WIDTH; x2 += stepSize)
		for (y2 = 0; y2 <= SCREEN_HEIGHT; y2 += stepSize)
			blit (target, screen, x2+x, y2+y, x2+x, y2+y, blockSize, blockSize);
}

static void topDownMultiWipeFade (BITMAP *target) {
	int z, zz ;

	for (zz = 0; zz < 8; zz++) {
		int lineCount = 0;
		for (z = zz; z < SCREEN_HEIGHT; z += 8) {
			blit (target, screen, 0, z, 0, z, SCREEN_WIDTH, 1);
			if (lineCount == 10) {
				lineCount = 0;
				rest (1);
			}
			lineCount++;
		}
	}
}

static void topDownWipeFade (BITMAP *target) {
	int z;

	for (z = 0; z <= SCREEN_HEIGHT / 8; z++) {
		blit (target, screen, 0, z * 8, 0, z * 8, SCREEN_WIDTH, 8);
		rest (1);
	}
}

static void bottomUpWipeFade (BITMAP *target) {
	int z;

	for (z = SCREEN_HEIGHT / 8; z >= 0; z--) {
		blit (target, screen, 0, z * 8, 0, z * 8, SCREEN_WIDTH, 8);
		rest (1);
	}
}

static void rightLeftWipeFade (BITMAP *target) {
	int z;

	for (z = SCREEN_WIDTH / 8; z >= 0; z--) {
		blit (target, screen, z * 8, 0, z * 8, 0, 8, SCREEN_HEIGHT);
		rest (1);
	}
}

static void leftRightWipeFade (BITMAP *target) {
	int z;

	for (z = 0; z <= SCREEN_WIDTH / 8; z++) {
		blit (target, screen, z * 8, 0, z * 8, 0, 8, SCREEN_HEIGHT);
		rest (1);
	}
}

static void crossLinesFade (BITMAP *target) {
	int z;

	for (z = 0; z < (SCREEN_WIDTH/4+4) / 4; z++) {
		blit (target, screen, z * 8, 0, z * 8, 0, 4, SCREEN_HEIGHT);
		blit (target, screen, (SCREEN_WIDTH-1) - (z * 8), 0, (SCREEN_WIDTH-1) - (z * 8), 0, 4, SCREEN_HEIGHT);
		blit (target, screen, 0, (z * 6), 0, (z * 6), SCREEN_WIDTH, 4);
		blit (target, screen, 0, (SCREEN_HEIGHT - (z * 6)), 0, (SCREEN_HEIGHT - (z * 6)), SCREEN_WIDTH, 4);
		rest (1);
	}
	for (z = (SCREEN_WIDTH/4+4) / 4; z != 0; z--) {
		blit (target, screen, z * 8 + 4, 0, z * 8 + 4, 0, 4, SCREEN_HEIGHT);
		blit (target, screen, (SCREEN_WIDTH-1) - (z * 8) + 4, 0, (SCREEN_WIDTH-1) - (z * 8) + 4, 0, 4, SCREEN_HEIGHT);
		blit (target, screen, 0, (z * 6) + 4, 0, (z * 6) + 4, SCREEN_WIDTH, 4);
		blit (target, screen, 0, (SCREEN_HEIGHT - (z * 6)) + 4, 0, (SCREEN_HEIGHT - (z * 6)) + 4, SCREEN_WIDTH, 4);
		rest (1);
	}
}


/*****************************************************************************
changeHandleError

Used internally by change and quickChange.  Prints out error messages.
*****************************************************************************/
static void changeHandleError( ) {
	if (errorMessage) {
		perror( errorMessage );
		errorMessage = NULL;
	}
}


/*****************************************************************************
change

Selects and executes a random transition.
*****************************************************************************/
void change(BITMAP * target)
{
	static void (* const procs[])(BITMAP*) = {
		crossLinesFade,			spiralPixelFade,
		leftRightWipeFade,		rightLeftWipeFade,
		bottomUpWipeFade,		topDownWipeFade,
		topDownMultiWipeFade,
	} ;
	static const int cprocs = sizeof( procs ) / sizeof( procs[0] );

	show_mouse (NULL);
	changeHandleError( );

	(procs[rand() % cprocs])( target );

	clear_keybuf ();
}


/*****************************************************************************
quickChange

Executes a fast and simple transition.

QUESTION: Why does change call clear_keybuf, but this does not
*****************************************************************************/
void quickChange (BITMAP *target) {
	changeHandleError( );
	blit (target, screen, 0, 0, 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT);
	rest (1);
}

